/*
 * $Id: SkipTestOn.java,v 1.11 2012/02/20 22:00:32 oboehm Exp $
 *
 * Copyright (c) 2010 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 25.01.2010 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.annotation;

import static patterntesting.runtime.NullConstants.NULL_STRING;

import java.lang.annotation.*;

/**
 * You have a test which should be skipped on Linux or another operationg
 * system? Or a test which should be skipped on a special JavaVM?
 * Then you can use this annotation to skip a single test (put the
 * annotation in front of the test method) or all tests in this class (put
 * the annotation in front of the class).
 * <br/>
 * What is the difference to <code>@Broken</code>?
 * "Broken" means, the test does not work for the moment and should be
 * (temporarily) skipped. "SkipTestOn" means, this test is not constructed
 * for that platform and should be therefore skipped on it.
 *
 * @see Broken
 * @see RunTestOn
 * @author oliver
 * @since 1.0 (25.01.2010)
 */
@Documented
@Target({ElementType.METHOD, ElementType.TYPE})
@Retention(RetentionPolicy.RUNTIME)
public @interface SkipTestOn {

    /**
     * Here you can define the name of the operating system or version of the
     * JDK for which a test should be skipped. But it is better to use the
     * attributes "osName" or "javaVersion" here because otherwise
     * PatternTesting tries to guess what you mean.
     * <br/>
     * Since 1.2 wildcards (*, ?) are supported.
     *
     * @see #osName()
     * @see #javaVersion()
     */
    String[] value() default NULL_STRING;

    /**
     * Should a test be skipped on a single platform e.g. for Linux?
     * Use "Linux" to mark it as broken only for the Linux platform.
     * Instead of "Linux" you can use any other operation system. The
     * format must be the same as returned by the system property "os.name".
     * Valid values are:
     * <ul>
     *  <li>"Linux"</li>
     *  <li>"Mac OS X" (for Mac),</li>
     *  <li>"Windows XP"</li>
     *  <li>and others
     *      (see {@link "http://lopica.sourceforge.net/os.html"}).
     *  </li>
     * </ul>
     * Multiple values like <code>{ "Linux", "Mac OS X" }</code> are allowed.
     * If no operation system is given you will get an IllegalArgumentException.
     * <br/>
     * The format of this attribute must be the same as returned by system
     * property "os.name".
     * <br/>
     * Since 1.2 wildcards (*, ?) are supported.
     */
    String[] osName() default NULL_STRING;

    /**
     * You want the test only to be skipped for the Intel architecture?
     * Then you can use this attribute to limit it on this platform.
     * You can define a single platform like <code>"x86_64"</code> for an
     * Intel-Mac with 64 bit or multiple platform like
     * <code>{ "x86_32", "x86_64" }</code> for Intel Mac with 42 or 64 bit.
     * <br/>
     * The format of this attribute must be the same as returned by the system
     * property "os.arch".
     * <br/>
     * Since 1.2 wildcards (*, ?) are supported.
     */
    String[] osArch() default NULL_STRING;

    /**
     * You want the test only to be skipped for a special version of the
     * operating system? Use this attribute to limit it.
     * In contradiction to the other attributes the real version must start
     * with the given version. I.e. if you define "10.6" as version this would
     * match only real version "10.6.1" or "10.6.2".
     * <br/>
     * You can't define a range of skipped versions. But several versions are
     * allowed. So if you want to skip the tests in version 10.6.1 till 10.6.3
     * define { "10.6.1", "10.6.2", "10.6.3" } as values.
     * The format of this attribute must be the same as returned by the system
     * property "osVersion".
     * <br/>
     * Since 1.2 wildcards (*, ?) are supported.
     */
    String[] osVersion() default NULL_STRING;

    /**
     * With this attribute you can express that the test should be skipped on
     * the given host(s). You can define the host by its name or by its
     * IP address.
     * <br/>
     * Since 1.2 wildcards (*, ?) are supported.
     */
    String[] host() default NULL_STRING;

    /**
     * You want the test to be skipped on a special version of the VM, e.g. on
     * version "1.6.0_17" because you know that on this version there is a bug
     * that caused your test not to work? Then use this attribute.
     * <br/>
     * You can't define a range of skipped versions. But you regex (regular
     * expressions) are allowed. And you can define more than one version
     * (or regex).
     * <br/>
     * The format of this attribute must be the same as returned by the
     * system property "java.version".
     * <br/>
     * Since 1.2 wildcards (*, ?) are supported.
     */
    String[] javaVersion() default NULL_STRING;

    /**
     * You want the test to be skipped on a special vendor VM? Use this attribute
     * here. The format of this attribute must be the same as returned by the
     * system property "java.vendor".
     * Valid values are:
     * <ul>
     *  <li>"Apple Inc."</li>
     *  <li>"IBM Corporation"</li>
     *  <li>and others</li>
     * </ul>
     * <br/>
     * Since 1.2 wildcards (*, ?) are supported.
     */
    String[] javaVendor() default NULL_STRING;

    /**
     * Does the test should skipped only for user "Bob" because he is not allow
     * to use the database?
     * Use <code>user="bob"</code> to skip it for his account.
     * <br/>
     * Damn, the new member of the team, Bill, has similar restrictions.
     * Ok, use <code>user={"bob", "bill"}</code> skip the test for both
     * accounts.
     * <br/>
     * Since 1.2 wildcards (*, ?) are supported.
     *
     * @since 1.1
     */
    String[] user() default NULL_STRING;

    /**
     * You have some tests that does not run when you are offline? Define a
     * property (e.g. <code>property="offline"</code>).
     * If this property is set (and its value is
     * not "false") the test will be skipped if you start the JavaVM with
     * <code>java -Doffline ...</code> or <code>java -Doffline=true ...</code>.
     * <br/>
     * You can control more than property with this attribute. Then all
     * properties must be "true" to skip the test.
     * <br/>
     * Wildcards (*, ?) are not yet supported for system properties.
     */
    String[] property() default NULL_STRING;

}

