/*
 * $Id: ConnectionMonitor.java,v 1.4 2013/01/02 17:54:38 oboehm Exp $
 *
 * Copyright (c) 2012 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 07.10.2012 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.db;

import java.sql.Connection;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import javax.management.JMException;

import org.slf4j.*;

import patterntesting.runtime.jmx.MBeanHelper;


/**
 * This is the monitor class for the {@link ProxyConnection} which monitors the
 * different newInstance() and close() call. So you can ask this class how many
 * and which connections are open.
 * <p>
 * The instance of this class is automatically registered as MBean as soon as a
 * {@link ProxyConnection} is used.
 * </p>
 *
 * @author oliver (ob@aosd.de)
 * @since 1.3 (07.10.2012)
 * @version $Revision: 1.4 $
 */
public final class ConnectionMonitor implements ConnectionMonitorMBean {

    private static final Logger log = LoggerFactory.getLogger(ConnectionMonitor.class);
    private static final ConnectionMonitorMBean instance;
    private static final Set<ProxyConnection> openConnections = new CopyOnWriteArraySet<ProxyConnection>();

    static {
        instance = new ConnectionMonitor();
        log.debug("{} created and registered as MBean.", instance);
        try {
            MBeanHelper.registerMBean(instance);
        } catch (JMException e) {
            log.info("{} can't be registered as MBean ({})", instance, e);
        }
    }

    /** No need to instantiate it - we provide only some services. */
    private ConnectionMonitor() {
        log.trace("{} + created.", this);
    }

    /**
     * Yes, it is a Singleton because it offers only some services. So we don't
     * need the object twice.
     *
     * @return the only instance
     */
    public static ConnectionMonitorMBean getInstance() {
        return instance;
    }

    /**
     * If you want to monitor the connection use this method.
     *
     * @param connection the original connection
     * @return a proxy for the connection
     */
    public static Connection getMonitoredConnection(final Connection connection) {
        return ProxyConnection.newInstance(connection);
    }

    /**
     * Adds the connection.
     *
     * @param proxyConnection the proxy connection
     */
    public static void addConnection(final ProxyConnection proxyConnection) {
        openConnections.add(proxyConnection);
    }

    /**
     * Removes the connection.
     *
     * @param proxyConnection the proxy connection
     */
    public static void removeConnection(final ProxyConnection proxyConnection) {
        openConnections.remove(proxyConnection);
    }

    /**
     * Gets the caller of the given connection.
     *
     * @param connection the connection
     * @return the caller of
     */
    public static StackTraceElement getCallerOf(final Connection connection) {
        for (ProxyConnection proxy : openConnections) {
            if (proxy.getConnection().equals(connection)) {
                return proxy.getCaller()[0];
            }
        }
        throw new IllegalArgumentException("not monitored or closed: " + connection);
    }

    /**
     * Gets the callers.
     *
     * @return the callers
     * @see patterntesting.runtime.db.ConnectionMonitorMBean#getCallers()
     */
    public StackTraceElement[] getCallers() {
        StackTraceElement[] callers = new StackTraceElement[openConnections.size()];
        int i = 0;
        for (ProxyConnection proxy : openConnections) {
            callers[i] = proxy.getCaller()[0];
        }
        return callers;
    }

    /**
     * Gets the open count.
     *
     * @return the open count
     * @see patterntesting.runtime.db.ConnectionMonitorMBean#getCount()
     */
    public int getCount() {
        return ConnectionMonitor.openConnections.size();
    }

    /**
     * Assert that all connections are closed.
     */
    public static void assertConnectionsClosed() {
        int count = instance.getCount();
        if (count > 0) {
            AssertionError error = new AssertionError(count + " connection(s) not closed");
            error.setStackTrace(openConnections.iterator().next().getCaller());
            throw error;
        }
    }

}

