/*
 * $Id: ProxyConnection.java,v 1.5 2012/10/07 17:50:46 oboehm Exp $
 *
 * Copyright (c) 2012 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 29.09.2012 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.db;

import java.lang.reflect.*;
import java.sql.Connection;

import org.slf4j.*;

/**
 * This is a dynamic proxy for a JDBC connection which monitors together with
 * the {@link ConnectionMonitor} the different newInstance() and close() call.
 *
 * @author oliver (ob@aosd.de)
 * @since 1.3 (29.09.2012)
 * @version $Revision: 1.5 $
 */
public final class ProxyConnection implements InvocationHandler {

    private static final Logger log = LoggerFactory.getLogger(ProxyConnection.class);
    private final Connection connection;
    private final StackTraceElement[] caller;

    /**
     * Creates a new proxy instance for the given connection.
     *
     * @param connection the connection
     * @return the connection
     */
    public static Connection newInstance(Connection connection) {
        ProxyConnection proxyConnection = new ProxyConnection(connection);
        ConnectionMonitor.addConnection(proxyConnection);
        Class<?>[] interfaces = new Class[] { Connection.class };
        return (Connection) Proxy.newProxyInstance(connection.getClass().getClassLoader(),
                interfaces, proxyConnection);
    }

    private ProxyConnection(final Connection connection) {
        this.connection = connection;
        this.caller = getCallerStacktrace(4);
    }

    private static StackTraceElement[] getCallerStacktrace(final int callsIgnored) {
        StackTraceElement[] stacktrace = Thread.currentThread().getStackTrace();
        StackTraceElement[] stacktraceCaller = new StackTraceElement[stacktrace.length - callsIgnored];
        System.arraycopy(stacktrace, callsIgnored, stacktraceCaller, 0, stacktrace.length - callsIgnored);
        return stacktraceCaller;
    }

    /**
     * Invoke.
     *
     * @param proxy the proxy
     * @param method the method
     * @param args the args
     * @return the object
     * @throws Throwable the throwable
     * @see java.lang.reflect.InvocationHandler#invoke(Object, Method, Object[])
     */
    public Object invoke(final Object proxy, final Method method, final Object[] args) throws Throwable {
        assert method != null;
        try {
            if (method.getName().equals("close")) {
                ConnectionMonitor.removeConnection(this);
            } else if (method.getName().equals("toString")) {
                return this.toString();
            }
            return method.invoke(connection, args);
        } catch (InvocationTargetException ite) {
            log.debug("Cannot invoke {} ({})!", method, ite);
            throw ite.getTargetException();
        }
    }
    
    /**
     * Gets the wrapped connection.
     *
     * @return the connection
     */
    public Connection getConnection() {
        return this.connection;
    }
    
    /**
     * Gets the stacktrace of the caller.
     *
     * @return the caller stacktrace
     */
    public StackTraceElement[] getCaller() {
        return this.caller;
    }

    /**
     * To string.
     *
     * @return the string
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return this.getClass().getSimpleName() + " for " + this.caller[0];
    }

}

